import argparse

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects import base_effect
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, argtypes


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "middleout",
        formatter_class=argtypes.CustomFormatter,
        help="Text expands in a single row or column in the middle of the output area then out.",
        description="Text expands in a single row or column in the middle of the output area then out.",
        epilog=f"""{argtypes.EASING_EPILOG}

Example: terminaltexteffects middleout -a 0.01 --expand-direction horizontal --center-movement-speed 0.35 
         --full-movement-speed 0.35 --center-easing IN_OUT_SINE --full-easing IN_OUT_SINE 
         --center-expand-color 00ff00 --full-expand-color 0000ff --starting-color ff0000""",
    )
    effect_parser.set_defaults(effect_class=MiddleoutEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time, in seconds, between animation steps.",
    )
    effect_parser.add_argument(
        "--starting-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the initial text in the center of the output area.",
    )
    effect_parser.add_argument(
        "--center-expand-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the row or column as it expands.",
    )
    effect_parser.add_argument(
        "--full-expand-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the full text as it expands.",
    )
    effect_parser.add_argument(
        "--expand-direction",
        default="vertical",
        choices=["vertical", "horizontal"],
        help="Direction the text will expand.",
    )
    effect_parser.add_argument(
        "--center-movement-speed",
        type=argtypes.valid_speed,
        default=0.35,
        metavar="(float > 0)",
        help="Speed of the characters during the initial expansion of the center vertical/horiztonal line. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--full-movement-speed",
        type=argtypes.valid_speed,
        default=0.35,
        metavar="(float > 0)",
        help="Speed of the characters during the final full expansion. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--center-easing",
        default="IN_OUT_SINE",
        type=argtypes.valid_ease,
        help="Easing function to use for initial expansion.",
    )
    effect_parser.add_argument(
        "--full-easing",
        default="IN_OUT_SINE",
        type=argtypes.valid_ease,
        help="Easing function to use for full expansion.",
    )


class MiddleoutEffect(base_effect.Effect):
    """Effect that ___."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        super().__init__(terminal, args)
        self.center_gradient = graphics.Gradient(self.args.starting_color, self.args.center_expand_color, 10)
        self.full_gradient = graphics.Gradient(self.args.center_expand_color, self.args.full_expand_color, 10)

    def prepare_data(self) -> None:
        """Prepares the data for the effect."""
        for character in self.terminal.characters:
            character.motion.set_coordinate(
                self.terminal.output_area.center_column, self.terminal.output_area.center_row
            )
            # setup scenes
            for step in self.center_gradient:
                character.animation.add_effect_to_scene("center", character.input_symbol, step, 2)
            for step in self.full_gradient:
                character.animation.add_effect_to_scene("full", character.input_symbol, step, 2)
            character.animation.scenes["full"].waypoint_sync_id = "full"
            character.animation.activate_scene("center")

            # setup waypoints
            if self.args.expand_direction == "vertical":
                column = character.input_coord.column
                row = self.terminal.output_area.center_row
            else:
                column = self.terminal.output_area.center_column
                row = character.input_coord.row
            character.motion.new_waypoint(
                "center",
                column,
                row,
                speed=self.args.center_movement_speed,
                ease=self.args.center_easing,
            )
            character.motion.new_waypoint(
                "full",
                character.input_coord.column,
                character.input_coord.row,
                speed=self.args.full_movement_speed,
                ease=self.args.full_easing,
            )
            # initialize character state
            character.motion.activate_waypoint("center")
            character.animation.activate_scene("center")
            character.is_active = True
            self.animating_chars.append(character)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        final = False
        while self.pending_chars or self.animating_chars:
            if all([character.motion.active_waypoint is None for character in self.animating_chars]):
                final = True
                for character in self.animating_chars:
                    character.motion.activate_waypoint("full")
                    character.animation.activate_scene("full")
            self.terminal.print()
            self.animate_chars()
            if final:
                self.animating_chars = [
                    animating_char
                    for animating_char in self.animating_chars
                    if not animating_char.animation.active_scene_is_complete()
                    or not animating_char.motion.movement_is_complete()
                ]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()
            animating_char.animation.step_animation()



import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects import base_character, base_effect
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, argtypes


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "unstable",
        formatter_class=argtypes.CustomFormatter,
        help="Spawn characters jumbled, explode them to the edge of the output area, then reassemble them in the correct layout.",
        description="unstable | Spawn characters jumbled, explode them to the edge of the output area, then reassemble them in the correct layout.",
        epilog=f"""{argtypes.EASING_EPILOG}

Example: terminaltexteffects unstable -a 0.01 --initial-color ffffff --unstable-color ff9200 --final-color ffffff --explosion-ease OUT_EXPO --explosion-speed 0.75 --reassembly-ease OUT_EXPO --reassembly-speed 0.75""",
    )
    effect_parser.set_defaults(effect_class=UnstableEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time, in seconds, between animation steps.",
    )
    effect_parser.add_argument(
        "--initial-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the characters when the effect starts.",
    )
    effect_parser.add_argument(
        "--unstable-color",
        type=argtypes.valid_color,
        default="ff9200",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color transitioned to as the characters become unstable.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color transitioned to as the characters reassemble.",
    )
    effect_parser.add_argument(
        "--explosion-ease",
        default="OUT_EXPO",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement during the explosion.",
    )
    effect_parser.add_argument(
        "--explosion-speed",
        type=argtypes.valid_speed,
        default=0.75,
        metavar="(float > 0)",
        help="Speed of characters during explosion. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--reassembly-ease",
        default="OUT_EXPO",
        type=argtypes.valid_ease,
        help="Easing function to use for character reassembly.",
    )
    effect_parser.add_argument(
        "--reassembly-speed",
        type=argtypes.valid_speed,
        default=0.75,
        metavar="(float > 0)",
        help="Speed of characters during reassembly. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )


class UnstableEffect(base_effect.Effect):
    """Effect that spawns characters jumbled, explodes them to the edge of the output area,
    then reassembles them in the correct layout."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        super().__init__(terminal, args)
        self.jumbled_coords: dict[base_character.EffectCharacter, base_character.motion.Coord] = dict()

    def prepare_data(self) -> None:
        """Prepares the data for the effect by jumbling the character positions and
        choosing a location on the perimeter of the output area for the character to travel
        after exploding. Creates all waypoints and scenes for the characters."""
        character_coords = [character.input_coord for character in self.terminal.characters]
        for character in self.terminal.characters:
            pos = random.randint(0, 3)
            match pos:
                case 0:
                    col = self.terminal.output_area.left
                    row = random.randint(1, self.terminal.output_area.top)
                case 1:
                    col = self.terminal.output_area.right
                    row = random.randint(1, self.terminal.output_area.top)
                case 2:
                    col = random.randint(1, self.terminal.output_area.right)
                    row = self.terminal.output_area.bottom
                case 3:
                    col = random.randint(1, self.terminal.output_area.right)
                    row = self.terminal.output_area.top
            jumbled_coord = character_coords.pop(random.randint(0, len(character_coords) - 1))
            self.jumbled_coords[character] = jumbled_coord
            character.motion.set_coordinate(jumbled_coord.column, jumbled_coord.row)
            character.motion.new_waypoint("explosion", col, row, speed=0.75, ease=self.args.explosion_ease)
            character.motion.new_waypoint(
                "reassembly",
                character.input_coord.column,
                character.input_coord.row,
                speed=0.75,
                ease=self.args.reassembly_ease,
            )
            unstable_gradient = graphics.Gradient(self.args.initial_color, self.args.unstable_color, 25)
            for step in unstable_gradient:
                character.animation.add_effect_to_scene("rumble", character.input_symbol, step, 10)
            final_color = graphics.Gradient(self.args.unstable_color, self.args.final_color, 12)
            for step in final_color:
                character.animation.add_effect_to_scene("final", character.input_symbol, step, 5)
            character.animation.activate_scene("rumble")
            character.is_active = True

    def move_all_to_waypoint(self, waypoint_id) -> None:
        for character in self.terminal.characters:
            if waypoint_id == "reassembly":
                character.animation.activate_scene("final")
            self.animating_chars.append(character)
            character.motion.activate_waypoint(waypoint_id)
        while self.animating_chars:
            self.terminal.print()
            self.animate_chars()
            if waypoint_id == "reassembly":
                self.animating_chars = [
                    animating_char
                    for animating_char in self.animating_chars
                    if not animating_char.motion.current_coord == animating_char.motion.waypoints[waypoint_id].coord
                    or not animating_char.animation.active_scene_is_complete()
                ]
            else:
                self.animating_chars = [
                    animating_char
                    for animating_char in self.animating_chars
                    if not animating_char.motion.current_coord == animating_char.motion.waypoints[waypoint_id].coord
                ]

    def rumble(self) -> None:
        max_rumble_steps = 250
        current_rumble_steps = 0
        rumble_mod_delay = 20
        while current_rumble_steps < max_rumble_steps:
            if current_rumble_steps > 30 and current_rumble_steps % rumble_mod_delay == 0:
                row_offset = random.choice([-1, 0, 1])
                column_offset = random.choice([-1, 0, 1])
                for character in self.terminal.characters:
                    character.motion.set_coordinate(
                        character.motion.current_coord.column + column_offset,
                        character.motion.current_coord.row + row_offset,
                    )
                    character.animation.step_animation()
                self.terminal.print()
                for character in self.terminal.characters:
                    character.motion.set_coordinate(
                        self.jumbled_coords[character].column, self.jumbled_coords[character].row
                    )
                rumble_mod_delay -= 1
                rumble_mod_delay = max(rumble_mod_delay, 1)
            else:
                for character in self.terminal.characters:
                    character.animation.step_animation()
                self.terminal.print()

            current_rumble_steps += 1

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        explosion_hold_time = 50
        self.rumble()
        self.move_all_to_waypoint("explosion")
        while explosion_hold_time:
            self.terminal.print()
            self.animate_chars()
            explosion_hold_time -= 1
        self.move_all_to_waypoint("reassembly")

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation."""
        for animating_char in self.animating_chars:
            animating_char.animation.step_animation()
            animating_char.motion.move()



import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects import base_character, base_effect
from terminaltexteffects.base_character import EventHandler
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, argtypes, motion, easing


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "bubbles",
        formatter_class=argtypes.CustomFormatter,
        help="Characters are formed into bubbles that float down and pop.",
        description="Characters are formed into bubbles that float down and pop.",
        epilog=f"""{argtypes.EASING_EPILOG}

Example: terminaltexteffects bubbles -a 0.01 --pop-color ff9600 --final-color 252525 --bubble-speed 0.1 --bubble-delay 50 --easing IN_OUT_SINE""",
    )
    effect_parser.set_defaults(effect_class=BubblesEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time, in seconds, between animation steps.",
    )
    effect_parser.add_argument(
        "--no-rainbow",
        action="store_true",
        help="If set, the bubbles will not be colored with a rainbow gradient.",
    )
    effect_parser.add_argument(
        "--bubble-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the bubbles. Ignored if --no-rainbow is left as default False.",
    )
    effect_parser.add_argument(
        "--pop-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the spray emitted when a bubble pops.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the final character.",
    )
    effect_parser.add_argument(
        "--bubble-speed",
        type=argtypes.valid_speed,
        default=0.1,
        metavar="(float > 0)",
        help="Speed of the floating bubbles. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--bubble-delay",
        type=argtypes.positive_int,
        default=50,
        metavar="(int > 0)",
        help="Number of animation steps between bubbles.",
    )
    effect_parser.add_argument(
        "--pop-condition",
        default="row",
        choices=["row", "bottom", "anywhere"],
        help="Condition for a bubble to pop. 'row' will pop the bubble when it reaches the the lowest row for which a character in the bubble originates. 'bottom' will pop the bubble at the bottom row of the terminal. 'anywhere' will pop the bubble randomly, or at the bottom of the terminal.",
    )
    effect_parser.add_argument(
        "--easing",
        default="IN_OUT_SINE",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement after a bubble pops.",
    )


class Bubble:
    def __init__(
        self,
        effect: "BubblesEffect",
        origin: base_character.motion.Coord,
        characters: list[base_character.EffectCharacter],
    ):
        self.effect = effect
        self.characters = characters
        self.radius = max(len(self.characters) // 5, 1)
        self.origin = origin
        self.anchor_char = base_character.EffectCharacter(" ", origin.column, origin.row, self.effect.terminal)
        if self.effect.args.pop_condition == "row":
            self.lowest_row = min([char.input_coord.row for char in self.characters])
        else:
            self.lowest_row = self.effect.terminal.output_area.bottom
        self.set_character_coordinates()
        self.landed = False
        self.make_waypoints()
        self.make_gradients()

    def set_character_coordinates(self) -> None:
        char: base_character.EffectCharacter
        point: base_character.motion.Coord
        for char, point in zip(
            self.characters,
            self.anchor_char.motion.find_points_on_circle(
                (self.anchor_char.motion.current_coord.column, self.anchor_char.motion.current_coord.row),
                self.radius,
                len(self.characters),
            ),
        ):
            char.motion.set_coordinate(point.column, point.row)
            if point.row == self.lowest_row:
                self.landed = True

        if self.effect.args.pop_condition == "anywhere":
            if random.random() < 0.002:
                self.landed = True

    def make_waypoints(self):
        waypoint_column = random.randint(self.effect.terminal.output_area.left, self.effect.terminal.output_area.right)
        self.anchor_char.motion.new_waypoint("floor", waypoint_column, self.lowest_row, self.effect.args.bubble_speed)
        self.anchor_char.motion.activate_waypoint("floor")

    def make_gradients(self) -> None:
        if self.effect.args.no_rainbow:
            for character in self.characters:
                character.animation.add_effect_to_scene("sheen", color=self.effect.args.bubble_color, duration=1)
                character.animation.activate_scene("sheen")
        else:
            rainbow_gradient = list(self.effect.rainbow_gradient.colors)
            gradient_offset = 0
            for character in self.characters:
                for step in rainbow_gradient:
                    character.animation.add_effect_to_scene("sheen", color=step, duration=5)
                gradient_offset += 2
                gradient_offset %= len(rainbow_gradient)
                rainbow_gradient = rainbow_gradient[gradient_offset:] + rainbow_gradient[:gradient_offset]
                character.animation.activate_scene("sheen")
                character.animation.active_scene.is_looping = True

    def pop(self) -> None:
        char: base_character.EffectCharacter
        point: base_character.motion.Coord
        for char, point in zip(
            self.characters,
            self.anchor_char.motion.find_points_on_circle(
                (self.anchor_char.motion.current_coord.column, self.anchor_char.motion.current_coord.row),
                self.radius + 3,
                len(self.characters),
            ),
        ):
            char.motion.new_waypoint("pop_out", point.column, point.row, 0.2, ease=easing.Ease["OUT_EXPO"], layer=1)
            char.event_handler.register_event(
                EventHandler.Event.WAYPOINT_REACHED, "pop_out", EventHandler.Action.ACTIVATE_WAYPOINT, "final"
            )
        for character in self.characters:
            character.animation.activate_scene("pop_1")
            character.motion.activate_waypoint("pop_out")

    def activate(self) -> None:
        for char in self.characters:
            char.is_active = True

    def move(self) -> None:
        self.anchor_char.motion.move()
        self.set_character_coordinates()
        for character in self.characters:
            character.animation.step_animation()


class BubblesEffect(base_effect.Effect):
    """Effect that ___."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        super().__init__(terminal, args)
        self.bubbles: list[Bubble] = []
        red = "e81416"
        orange = "ffa500"
        yellow = "faeb36"
        green = "79c314"
        blue = "487de7"
        indigo = "4b369d"
        violet = "70369d"
        red_orange = graphics.Gradient(red, orange, 5)
        orange_yellow = graphics.Gradient(orange, yellow, 5)
        yellow_green = graphics.Gradient(yellow, green, 5)
        green_blue = graphics.Gradient(green, blue, 5)
        blue_indigo = graphics.Gradient(blue, indigo, 5)
        indigo_violet = graphics.Gradient(indigo, violet, 5)
        violet_red = graphics.Gradient(violet, red, 5)
        self.rainbow_gradient = (
            red_orange + orange_yellow + yellow_green + green_blue + blue_indigo + indigo_violet + violet_red
        )

    def prepare_data(self) -> None:
        """Prepares the data for the effect by ___."""
        final_gradient = graphics.Gradient(self.args.pop_color, self.args.final_color, 10)
        for character in self.terminal.characters:
            character.motion.layer = 1
            character.animation.add_effect_to_scene("pop_1", "*", color=self.args.pop_color, duration=25)
            character.animation.add_effect_to_scene("pop_2", "'", color=self.args.pop_color, duration=25)
            for step in final_gradient:
                character.animation.add_effect_to_scene("final", color=step, duration=5)
            character.event_handler.register_event(
                EventHandler.Event.SCENE_COMPLETE, "pop_1", EventHandler.Action.ACTIVATE_SCENE, "pop_2"
            )
            character.event_handler.register_event(
                EventHandler.Event.SCENE_COMPLETE, "pop_2", EventHandler.Action.ACTIVATE_SCENE, "final"
            )
            character.motion.new_waypoint(
                "final",
                character.input_coord.column,
                character.input_coord.row,
                0.3,
                ease=easing.Ease["IN_OUT_EXPO"],
                layer=1,
            )

        unbubbled_chars = []
        for _, char_list in self.input_by_row().items():
            unbubbled_chars.extend(char_list)
        self.bubbles = []
        while unbubbled_chars:
            bubble_group = []
            if len(unbubbled_chars) < 5:
                bubble_group.extend(unbubbled_chars)
                unbubbled_chars.clear()
            else:
                for _ in range(random.randint(5, min(len(unbubbled_chars), 20))):
                    bubble_group.append(unbubbled_chars.pop(0))
            bubble_origin = motion.Coord(
                random.randint(self.terminal.output_area.left, self.terminal.output_area.right),
                self.terminal.output_area.top,
            )
            new_bubble = Bubble(self, bubble_origin, bubble_group)
            self.bubbles.append(new_bubble)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        animating_bubbles: list[Bubble] = []
        steps_since_last_bubble = 0
        while animating_bubbles or self.animating_chars or self.bubbles:
            if self.bubbles and steps_since_last_bubble >= self.args.bubble_delay:
                next_bubble = self.bubbles.pop(0)
                next_bubble.activate()
                animating_bubbles.append(next_bubble)
                steps_since_last_bubble = 0
            steps_since_last_bubble += 1

            for bubble in animating_bubbles:
                if bubble.landed:
                    bubble.pop()
                    self.animating_chars.extend(bubble.characters)

            animating_bubbles = [bubble for bubble in animating_bubbles if not bubble.landed]
            self.animate_bubbles(animating_bubbles)
            self.animate_chars()
            self.terminal.print()

            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
                or not animating_char.motion.movement_is_complete()
            ]

    def animate_bubbles(self, animating_bubbles: list[Bubble]) -> None:
        for bubble in animating_bubbles:
            bubble.move()

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation."""
        for animating_char in self.animating_chars:
            animating_char.animation.step_animation()
            animating_char.motion.move()



import typing
import random
from dataclasses import dataclass, field

from terminaltexteffects.utils import ansitools, colorterm, hexterm

if typing.TYPE_CHECKING:
    from terminaltexteffects import base_character


@dataclass
class Gradient:
    """A Gradient is a list of RGB hex color strings transitioning from one color to another. The gradient color
    list is calculated using linear interpolation based on the provided start and end colors and the number of steps.

    Args:
        start_color (str|int): RGB hex color string or XTerm-256 color code
        end_color (str|int): RGB hex color string or XTerm-256 color code
        steps (int): Number of intermediate colors to calculate

    Attributes:
        colors (list[str]): List (length=steps + 2) of RGB hex color strings

    """

    start_color: str | int
    end_color: str | int
    steps: int

    def __post_init__(self) -> None:
        self.colors: list[str] = self._generate()
        self.index: int = 0

    def _generate(self) -> list[str]:
        """Calculate a gradient of colors between two colors using linear interpolation.

        Returns:
            list[str]: List (length=steps + 2) of RGB hex color strings
        """
        # Convert start_color to hex if it's an XTerm-256 color code
        if isinstance(self.start_color, int):
            self.start_color = hexterm.xterm_to_hex(self.start_color)
        # Convert end_color to hex if it's an XTerm-256 color code
        if isinstance(self.end_color, int):
            self.end_color = hexterm.xterm_to_hex(self.end_color)
        # Convert start_color to a list of RGB values
        start_color_ints = colorterm._hex_to_int(self.start_color)
        # Convert end_color to a list of RGB values
        end_color_ints = colorterm._hex_to_int(self.end_color)
        # Initialize an empty list to store the gradient colors
        gradient_colors: list[str] = []
        # Add the start color to the gradient colors list
        gradient_colors.append(self.start_color)
        # Calculate the color deltas for each RGB value
        red_delta = (end_color_ints[0] - start_color_ints[0]) // self.steps
        green_delta = (end_color_ints[1] - start_color_ints[1]) // self.steps
        blue_delta = (end_color_ints[2] - start_color_ints[2]) // self.steps
        # Calculate the intermediate colors and add them to the gradient colors list
        for i in range(self.steps):
            red = start_color_ints[0] + (red_delta * i)
            green = start_color_ints[1] + (green_delta * i)
            blue = start_color_ints[2] + (blue_delta * i)

            # Ensure that the RGB values are within the valid range of 0-255
            red = max(0, min(red, 255))
            green = max(0, min(green, 255))
            blue = max(0, min(blue, 255))

            # Convert the RGB values to a hex color string and add it to the gradient colors list
            gradient_colors.append(f"{red:02x}{green:02x}{blue:02x}")
        # Add the end color to the gradient colors list
        gradient_colors.append(self.end_color)
        # Return the list of gradient colors
        return gradient_colors

    def __iter__(self) -> "Gradient":
        self.index = 0
        return self

    def __next__(self) -> str:
        if self.index < len(self.colors):
            color = self.colors[self.index]
            self.index += 1
            return color
        else:
            raise StopIteration

    def __add__(self, other: "Gradient") -> "Gradient":
        if not isinstance(other, Gradient):
            return NotImplemented
        new_gradient = Gradient("ffffff", "ffffff", 1)
        new_gradient.colors = self.colors + other.colors
        new_gradient.steps = self.steps + other.steps
        new_gradient.start_color = self.start_color
        new_gradient.end_color = other.end_color
        return new_gradient


@dataclass
class GraphicalEffect:
    """A class for storing terminal graphical modes and a color.

    Supported graphical modes:
    bold, dim, italic, underline, blink, inverse, hidden, strike

    Args:
        symbol (str): the symbol to show
        bold (bool): bold mode
        dim (bool): dim mode
        italic (bool): italic mode
        underline (bool): underline mode
        blink (bool): blink mode
        reverse (bool): reverse mode
        hidden (bool): hidden mode
        strike (bool): strike mode
        color (int | str): color code
    """

    symbol: str
    bold: bool = False
    dim: bool = False
    italic: bool = False
    underline: bool = False
    blink: bool = False
    reverse: bool = False
    hidden: bool = False
    strike: bool = False
    color: int | str | None = None

    def __post_init__(self):
        self.format_symbol()

    def disable_modes(self) -> None:
        """Disables all graphical modes."""
        self.bold = False
        self.dim = False
        self.italic = False
        self.underline = False
        self.blink = False
        self.reverse = False
        self.hidden = False
        self.strike = False

    def format_symbol(self) -> None:
        """Formats the symbol for printing."""
        formatting_string = ""
        if self.bold:
            formatting_string += ansitools.APPLY_BOLD()
        if self.italic:
            formatting_string += ansitools.APPLY_ITALIC()
        if self.underline:
            formatting_string += ansitools.APPLY_UNDERLINE()
        if self.blink:
            formatting_string += ansitools.APPLY_BLINK()
        if self.reverse:
            formatting_string += ansitools.APPLY_REVERSE()
        if self.hidden:
            formatting_string += ansitools.APPLY_HIDDEN()
        if self.strike:
            formatting_string += ansitools.APPLY_STRIKETHROUGH()
        if self.color is not None:
            formatting_string += colorterm.fg(self.color)

        self.symbol = f"{formatting_string}{self.symbol}{ansitools.RESET_ALL() if formatting_string else ''}"


@dataclass
class Sequence:
    """An Sequence is a graphicaleffect with a duration.
    Args:
        graphical_effect (GraphicalEffect): a GraphicalEffect object containing the graphical modes and color of the character
        duration (int): the number of animation steps to use the Sequence
    """

    graphical_effect: GraphicalEffect
    duration: int

    def __post_init__(self):
        self.frames_played = 0
        self.symbol = self.graphical_effect.symbol


@dataclass
class Scene:
    """A Scene is a list of Sequences.

    Args:
        id (str): the id of the Scene
        sequences (list[Sequence]): a list of Sequence objects
        is_looping (bool): whether the Scene should loop
    """

    id: str
    sequences: list[Sequence] = field(default_factory=list)
    is_looping: bool = False
    waypoint_sync_id: str | None = None

    def __post_init__(self):
        self.played_sequences = []

    def add_sequence(self, graphicaleffect: GraphicalEffect, duration: int) -> None:
        """Adds a Sequence to the Scene.

        Args:
            graphicaleffect (GraphicalEffect): a GraphicalEffect object containing the graphical modes and color of the character
            duration (int): the number of animation steps to use the Sequence
        """
        self.sequences.append(Sequence(graphicaleffect, duration))

    def activate(self) -> str:
        """Activates the Scene.

        Returns:
            str: the next symbol in the Scene
        """
        if self.sequences:
            return self.sequences[0].symbol
        else:
            raise ValueError("Scene has no sequences.")

    def apply_gradient(self, start_color: str | int, end_color: str | int) -> None:
        """
        Applies a gradient effect across the sequences of the scene. The gradient is generated
        to have the same number of steps as the number of sequences in the scene.

        Parameters
        ----------
        start_color : str | int
            The starting color of the gradient.
        end_color : str | int
            The ending color of the gradient.

        Returns
        -------
        None
        """
        if not self.sequences:
            return
        else:
            gradient = Gradient(start_color, end_color, len(self.sequences))
            for i, sequence in enumerate(self.sequences):
                sequence.graphical_effect.color = gradient.colors[i]
                sequence.graphical_effect.format_symbol()

    def get_next_symbol(self) -> str:
        """Returns the next symbol in the Scene.

        Returns:
            str: the next symbol in the Scene
        """

        current_sequence = self.sequences[0]
        next_symbol = current_sequence.symbol
        current_sequence.frames_played += 1
        if current_sequence.frames_played == current_sequence.duration:
            current_sequence.frames_played = 0
            self.played_sequences.append(self.sequences.pop(0))
            if self.is_looping and not self.sequences:
                self.sequences.extend(self.played_sequences)
                self.played_sequences.clear()
        return next_symbol

    def reset_scene(self) -> None:
        """Resets the Scene."""
        for sequence in self.sequences:
            sequence.frames_played = 0
            self.played_sequences.append(sequence)
        self.sequences.clear()
        self.sequences.extend(self.played_sequences)
        self.played_sequences.clear()

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, Scene):
            return NotImplemented
        return self.id == other.id


class Animation:
    def __init__(self, character: "base_character.EffectCharacter"):
        """Animation handles the animations of a character. It contains a list of Scenes, and the active scene id. GraphicalEffects are
        added to the Scenes, and the Animation returns the next symbol in the active scene.

        Args:
            character (base_character.EffectCharacter): the EffectCharacter object to animate
        """
        self.character = character
        self.scenes: dict[str, Scene] = {}
        self.active_scene: Scene | None = None
        self.use_xterm_colors: bool = False
        self.no_color: bool = False
        self.xterm_color_map: dict[str, int] = {}

    def add_effect_to_scene(
        self,
        scene_id: str,
        symbol: str | None | GraphicalEffect = None,
        color: int | str | None = None,
        duration: int = 1,
    ) -> "Scene":
        """Add a graphical effect to a scene. If the scene doesn't exist, it will be created.

        Args:
            scene_id (str): ID of the scene to which the effect will be added
            symbol (str | GraphicalEffect): Symbol to display, if None, the character's input symbol will be used. If a GraphicalEffect is passed, the color and graphical modes will be used.
            color (int | str): XTerm color code (0-255) or hex color code (e.g. ffffff)
            duration (int): Number of animation steps to display the effect, defaults to 1. Minimum 1.
        """
        if duration < 1:
            raise ValueError("duration must be greater than 0")
        if scene_id not in self.scenes:
            new_scene = Scene(scene_id)
            self.scenes[scene_id] = new_scene
        if not symbol:
            symbol = self.character.input_symbol
        if not self.no_color:
            if self.use_xterm_colors and isinstance(color, str):
                if color in self.xterm_color_map:
                    color = self.xterm_color_map[color]
                else:
                    xterm_color = hexterm.hex_to_xterm(color)
                    self.xterm_color_map[color] = xterm_color
                    color = xterm_color
        else:
            color = None
        if isinstance(symbol, str):
            graphicaleffect = GraphicalEffect(symbol, color=color)
        else:
            graphicaleffect = symbol
        self.scenes[scene_id].add_sequence(graphicaleffect, duration)
        return self.scenes[scene_id]

    def active_scene_is_complete(self) -> bool:
        """Returns whether the active scene is complete. A scene is complete if all sequences have been played.
        Looping scenes are always complete.

        Returns:
            bool: True if complete, False otherwise
        """
        if not self.active_scene:
            return True
        elif not self.active_scene.sequences or self.active_scene.is_looping:
            return True
        elif self.active_scene.waypoint_sync_id is not None:
            if (
                self.character.motion.active_waypoint is None
                or self.character.motion.active_waypoint.id != self.active_scene.waypoint_sync_id
            ):
                return True
        return False

    def random_color(self) -> str:
        """Returns a random hex color code.

        Returns:
            str: hex color code
        """
        return hex(random.randint(0, 0xFFFFFF))[2:].zfill(6)

    def step_animation(self) -> None:
        """Apply the next symbol in the scene to the character. If a scene order exists, the next scene
        will be activated when the current scene is complete."""
        if self.active_scene and self.active_scene.sequences:
            # if the active scene is synced to the active waypoint, calculate the sequence index based on the
            # current waypoint progress
            if self.active_scene.waypoint_sync_id is not None:
                if (
                    self.character.motion.active_waypoint
                    and self.character.motion.active_waypoint.id == self.active_scene.waypoint_sync_id
                ):
                    sequence_index = round(
                        len(self.active_scene.sequences)
                        * (
                            max(self.character.motion.inter_waypoint_current_step, 1)
                            / max(self.character.motion.inter_waypoint_max_steps, 1)
                        )
                    )
                    try:
                        self.character.symbol = self.active_scene.sequences[sequence_index].symbol
                    except IndexError:
                        self.character.symbol = self.active_scene.sequences[-1].symbol
                else:  # when the active waypoint has been deactivated or changed, use the final symbol in the scene
                    self.character.symbol = self.active_scene.sequences[-1].symbol
            else:
                self.character.symbol = self.active_scene.get_next_symbol()
            if self.active_scene_is_complete():
                completed_scene_id = self.active_scene.id
                if not self.active_scene.is_looping:
                    self.active_scene.reset_scene()
                    self.active_scene = None

                self.character.event_handler.handle_event(
                    self.character.event_handler.Event.SCENE_COMPLETE, completed_scene_id
                )

    def reset_scene(self, scene_id: str) -> None:
        """Resets the Scene.

        Args:
            scene_name (str): the ID of the Scene to reset
        """
        if scene_id in self.scenes:
            self.scenes[scene_id].reset_scene()

    def activate_scene(self, scene_id: str) -> None:
        """Sets the active scene.

        Args:
            scene_id (str): the ID of the Scene to set as active
        """
        if scene_id in self.scenes:
            self.active_scene = self.scenes[scene_id]
            self.character.symbol = self.active_scene.activate()
            self.character.event_handler.handle_event(self.character.event_handler.Event.SCENE_ACTIVATED, scene_id)
        else:
            raise ValueError(f"Scene {scene_id} does not exist")

    def deactivate_scene(self, scene_id: str) -> None:
        """Deactivates a scene.

        Args:
            scene_id (str): the ID of the Scene to deactivate
        """
        if self.active_scene and self.active_scene.id == scene_id:
            self.active_scene = None



import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects import base_character, base_effect
from terminaltexteffects.base_character import EventHandler
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, argtypes


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "errorcorrect",
        formatter_class=argtypes.CustomFormatter,
        help="Some characters start in the wrong position and are corrected in sequence.",
        description="Some characters start in the wrong position and are corrected in sequence.",
        epilog=f"""{argtypes.EASING_EPILOG}

Example: terminaltexteffects errorcorrect -a 0.01 --error-pairs 12 --swap-delay 70 --error-color e74c3c --correct-color 45bf55 
         --final-color ffffff --movement-speed 0.5""",
    )
    effect_parser.set_defaults(effect_class=ErrorCorrectEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time, in seconds, between animation steps.",
    )
    effect_parser.add_argument(
        "--error-pairs",
        type=argtypes.positive_int,
        default=12,
        metavar="(int > 0)",
        help="Number of character pairs to swap.",
    )
    effect_parser.add_argument(
        "--swap-delay",
        type=argtypes.positive_int,
        default=70,
        metavar="(int > 0)",
        help="Number of animation steps between swaps.",
    )
    effect_parser.add_argument(
        "--error-color",
        type=argtypes.valid_color,
        default="e74c3c",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the characters that are in the wrong position.",
    )
    effect_parser.add_argument(
        "--correct-color",
        type=argtypes.valid_color,
        default="45bf55",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the characters once corrected, this is a gradient from error-color and fades to final-color.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the characters in their correct position.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.5,
        metavar="(float > 0)",
        help="Speed of the characters while moving to the correct position. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )


class ErrorCorrectEffect(base_effect.Effect):
    """Effect that swaps characters from an incorrect initial position to the correct position."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        super().__init__(terminal, args)
        self.swapped: list[tuple[base_character.EffectCharacter, base_character.EffectCharacter]] = []

    def prepare_data(self) -> None:
        """Prepares the data for the effect by swapping positions and generating animations and waypoints."""
        for character in self.terminal.characters:
            character.animation.add_effect_to_scene("spawn", character.input_symbol, self.args.final_color, 1)
            character.animation.activate_scene("spawn")
            character.is_active = True
        all_characters: list[base_character.EffectCharacter] = list(self.terminal.characters)
        correcting_gradient = graphics.Gradient(self.args.error_color, self.args.correct_color, 10)
        final_gradient = graphics.Gradient(self.args.correct_color, self.args.final_color, 10)
        block_symbol = "▓"
        block_wipe_start = ("▁", "▂", "▃", "▄", "▅", "▆", "▇", "█")
        block_wipe_end = ("▇", "▆", "▅", "▄", "▃", "▂", "▁")
        for _ in range(self.args.error_pairs):
            if len(all_characters) < 2:
                break
            char1 = all_characters.pop(random.randrange(len(all_characters)))
            char2 = all_characters.pop(random.randrange(len(all_characters)))
            char1.motion.set_coordinate(char2.input_coord.column, char2.input_coord.row)
            char1.motion.new_waypoint(
                "input_coord",
                char1.input_coord.column,
                char1.input_coord.row,
                self.args.movement_speed,
                layer=1,
            )
            char2.motion.set_coordinate(char1.input_coord.column, char1.input_coord.row)
            char2.motion.new_waypoint(
                "input_coord",
                char2.input_coord.column,
                char2.input_coord.row,
                self.args.movement_speed,
                layer=1,
            )
            self.swapped.append((char1, char2))
            for character in (char1, char2):
                for block in block_wipe_start:
                    character.animation.add_effect_to_scene("circles_start", block, self.args.error_color, 3)
                for block in block_wipe_end:
                    character.animation.add_effect_to_scene("circles_end", block, self.args.correct_color, 3)
                character.animation.add_effect_to_scene("initial", character.input_symbol, self.args.error_color, 1)
                character.animation.activate_scene("initial")
                for _ in range(10):
                    character.animation.add_effect_to_scene("error", block_symbol, self.args.error_color, 3)
                    character.animation.add_effect_to_scene("error", character.input_symbol, "ffffff", 3)
                for step in correcting_gradient:
                    character.animation.add_effect_to_scene("correcting", "█", step, 3)
                character.animation.scenes["correcting"].waypoint_sync_id = "input_coord"
                for step in final_gradient:
                    character.animation.add_effect_to_scene("final", character.input_symbol, step, 3)
                character.event_handler.register_event(
                    EventHandler.Event.SCENE_COMPLETE,
                    "circles_start",
                    EventHandler.Action.ACTIVATE_WAYPOINT,
                    "input_coord",
                )
                character.event_handler.register_event(
                    EventHandler.Event.SCENE_COMPLETE, "error", EventHandler.Action.ACTIVATE_SCENE, "circles_start"
                )
                character.event_handler.register_event(
                    EventHandler.Event.SCENE_COMPLETE, "circles_start", EventHandler.Action.ACTIVATE_SCENE, "correcting"
                )
                character.event_handler.register_event(
                    EventHandler.Event.WAYPOINT_REACHED,
                    "input_coord",
                    EventHandler.Action.ACTIVATE_SCENE,
                    "circles_end",
                )
                character.event_handler.register_event(
                    EventHandler.Event.SCENE_COMPLETE, "circles_end", EventHandler.Action.ACTIVATE_SCENE, "final"
                )

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        while self.swapped or self.animating_chars:
            swap_delay = self.args.swap_delay
            if self.swapped:
                next_pair = self.swapped.pop(0)
                for char in next_pair:
                    char.animation.activate_scene("error")
                    self.animating_chars.append(char)
            while swap_delay:
                self.terminal.print()
                self.animate_chars()

                self.animating_chars = [
                    animating_char
                    for animating_char in self.animating_chars
                    if not animating_char.animation.active_scene_is_complete()
                    or not animating_char.motion.movement_is_complete()
                ]
                if not self.animating_chars:
                    pass
                swap_delay -= 1
        self.terminal.print()

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation. Move characters prior to stepping animation
        to ensure waypoint synced animations have the latest waypoint progress information."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()
            animating_char.animation.step_animation()
